# -*- coding: utf-8 -*-

# Esiteks paneme thele, et lesande sisuks on Fibonacci jada arvutamine, kus on 
# etteantud kaks esimest liiget, ning iga jrgmine on kahe eelmise summa.

# Paneme thele ka seda, et me saame vtta tulemust moodulo K peale igat operatsiooni,
# mitte ainult lpus.

# Kige lihtsam lahendus on seega jrgmine otsearvutus keerukusega O(t)
def naive(N, V, T, K):
    (n,v) = N % K, V % K
    t = 0
    while t < T:
        n, v = v, (n + v) % K
        t += 1
    return (n + v) % K

# Paneme thele, et kik arvutava Fibonacci jada elemendid 
# on piiratud 0 <= x < K ning iga jrgmine on mratud kahe eelmise poolt.
# Seega jadas peab olema tskkel pikkusega kuni K^2,

# Enamgi, saab nidata et tskkel kindlasti peab algama kohe jada algusest,
# s.t. varem vi hiljem peavad esialgsed (n, v) vrtused kindlasti korduma, ning edasi
# lheb kogu protsess tskliliseks.
# Katsetades saab ka thele panna et enamasti on tskkel oluliselt lhem kui K^2, 
# seega antud lahendus lbib ka suhteliselt suurte K korral teste.

# Siinkohas paneme thele et sama lahendus C++-s saab ikkagi 10 punkti rohkem lihtsalt tnu
# kompileeritud koodi parema kiirusele.
def cycle(N, V, T, K):
    (n_start, v_start) = (n, v) = (N % K, V % K)
    t = 0
    while t < T:
        n, v = v, (n + v) % K
        t += 1
        if (n, v) == (n_start, v_start):
            # Leidsime tsklit pikkusega t, meil jb seega teha reaalselt ainult T % t sammu
            # Saame le hpata
            T = T % t
            t = 0
    return (n + v) % K


# Tegelikult muidugi saab nutud arvu vlja arvutada ka kinnise valemina, kui 
# vtta kasutusele natuke maatriksalgebrat
# Vaatame kuidas neb vlja leminek paarilt (n, v)-lt paarile (n', v') he sekundi prast:
#   n'    = 0*n + 1*v
#   v'    = 1*n + 1*v
# Seda saab kirja panna maatrikskorrutisena:
#   n'    = (0 1) * (n)
#   v'    = (1 1)   (v)
# Thistades maatriksi A := [[0, 1], [1, 1]] neme et T sekundi prast on vektor n_t, v_t:
#   n_t   =  (A^t)* (a_0)
#   v_t   =         (a_1)
# Maatriksi astendamist saab aga teha logaritmilisel ajal klassikalise astendamise algoritmi abil
def fast(N, V, T, K):
    n, v = N % K, V % K
    if T > 0:
        result = matrix_multiply(matrix_power([[0, 1], [1, 1]], T, K), [[n], [v]], K)
        n, v = result[0][0], result[1][0]
    return (n + v) % K

def matrix_multiply(A, B, K):
    return [[sum([A[r][k]*B[k][c] for k in range(len(B))]) % K 
            for c in range(len(B[0]))] for r in range(len(A))]

def matrix_power(A, n, K):
    if n == 1: return A
    A_halfn = matrix_power(A, n/2, K)
    A_halfn2 = matrix_multiply(A_halfn, A_halfn, K)
    if (n % 2 == 1): 
        return matrix_multiply(A_halfn2, A, K)
    else: 
        return A_halfn2